@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 from __future__ import annotations
 
-__version__ = "1.1.2"
+__version__ = "1.1.4"
 
 import glob
 import sys
@@ -24,10 +24,11 @@ import urllib.request
 from pathlib import Path
 from zipfile import ZipFile
 import inspect as _inspect
+import re
 
 from typing import TYPE_CHECKING
 if TYPE_CHECKING:
-    from typing import List, Dict, Tuple, Union, Optional
+    from typing import List, Dict, Tuple, Union, Optional, Any
 
 
 
@@ -132,8 +133,9 @@ def _is_git_url(url: str) -> bool:
 def _debug(*msgs) -> None:
     """ Print debug info only if debugging is turned on """
     if register.debug:
-        caller = _inspect.stack()[1][3]
-        print(f"DEBUG::{caller[:24]}".ljust(32), ":", *msgs, file=sys.stderr)
+        l = 28
+        caller = _abbrev(_inspect.stack()[1][3], l)
+        print(f"DEBUG:{caller.ljust(l)}:", *msgs, file=sys.stderr)
 
 
 def _errormsg(msg: str) -> None:
@@ -168,16 +170,18 @@ class Asset:
     The use case is for data files which are needed by a plugin at runtime
 
     Attribs:
-        url: the url where to download the assets from. It can be a git repository
-            or the direct url to a downloadable file/zip file
-        paths: the paths to extract from the url (if needed). Each item can be either
+        source: the source where to get the assets from. It can be a url to a git repository,
+            the direct url to a downloadable file/zip file or the path to a local
+            folder
+        patterns: the paths to extract from the source (if needed). Each item can be either
             a path relative to the root of the git or zip file, or a glob pattern. In the
             case where the url points to a concrete file (not a zip or a git repo), this
             attributes should be empty
         platform: if given, the platform for which this assets are valid (in the case
             of binaries of some kind)
+        name: the name of the asset (optional)
     """
-    url: str
+    source: str
     """the url where to download the assets from"""
 
     patterns: List[str] = None
@@ -188,23 +192,33 @@ class Asset:
 
     name: str = ''
 
+    def __post_init__(self):
+        assert self.source
+
     def identifier(self) -> str:
         if self.name:
             return self.name
         if self.patterns:
-            return f"{self.url}::{','.join(str(patt) for patt in self.patterns)}"
-        return self.url
+            return f"{self.source}::{','.join(str(patt) for patt in self.patterns)}"
+        return self.source
 
     def local_path(self) -> Path:
-        assert self.url
-        if _is_git_url(self.url):
-            return _git_local_path(self.url)
-        _debug(f"Downloading url {self.url}")
-        return _download_file(self.url)
+        assert self.source
+        if _is_url(self.source):
+            if _is_git_url(self.source):
+                return _git_local_path(self.source)
+            else:
+                _debug(f"Downloading url {self.source}")
+                return _download_file(self.source)
+        else:
+            # it is a path, check that it exists
+            source = Path(self.source)
+            assert source.exists(), f"Assert source does not exist: {source}"
+            return source
 
     def retrieve(self) -> List[Path]:
         """
-        Download and resolve all files
+        Download and resolve all files, if needed
 
         Returns:
             a (possibly empty) list of local paths which belong to this asset.
@@ -212,7 +226,8 @@ class Asset:
             files are extracted to a temp dir and a path to that temp dir
             is returned
         """
-        assert self.url and _is_url(self.url)
+        assert self.source and (_is_url(self.source) or os.path.isabs(self.source)), \
+            f"Source should be either a url or an absolute path: {self.source}"
         # self.url is either a git repo or a url pointing to a file
         root = self.local_path()
         if root.is_dir():
@@ -363,11 +378,6 @@ class Plugin:
     doc_folder: str = 'doc'
     assets: Optional[List[Asset]] = None
 
-    def __post_init__(self):
-        # manifest_relative_path should be either empty or a subdir of the repository's root. It should
-        # not point directly to the manifest, since this is hard-coded as risset.json
-        assert not self.manifest_relative_path or not os.path.splitext(self.manifest_relative_path)
-
     def __hash__(self):
         return hash((self.name, self.version))
 
@@ -420,6 +430,7 @@ class Opcode:
     name: str
     plugin: str
     syntaxes: Optional[List[str]] = None
+    installed: bool = True
 
 
 @dataclass
@@ -646,9 +657,13 @@ def _get_binary(binary) -> Optional[str]:
 
 
 def _get_git_binary() -> str:
+    cached = register.cache.get('git-binary')
+    if cached:
+        return cached
     path = shutil.which("git")
     if not path or not os.path.exists(path):
         raise RuntimeError("git binary not found")
+    register.cache['git-binary'] = path
     return path
 
 
@@ -704,13 +719,34 @@ def _git_clone_into(repo: str, destination: Path, depth=1) -> None:
     subprocess.call(args)
 
 
-def _git_update(repopath: Path, depth=0) -> None:
+def _git_repo_needs_update(repopath: Path) -> bool:
+    """
+    Check if a repository needs to be updated
+
+    NB: for our use case, where no merges are expected, to update is just
+    as fast as to check first and then act.
+    """
+    cwd = os.path.abspath(os.path.curdir)
+    os.chdir(str(repopath))
+    git = _get_git_binary()
+    subprocess.call([git, "fetch"])
+    headhash = subprocess.check_output([git, "rev-parse", "HEAD"])
+    upstreamhash =subprocess.check_output([git, "rev-parse", "master@{upstream}"])
+    _debug(f"Checking hashes, head: {headhash}, upstream: {upstreamhash}")
+    os.chdir(cwd)
+    return headhash != upstreamhash
+
+
+def _git_update(repopath: Path, depth=0, check_if_needed=False) -> None:
     """
     Update the git repo at the given path
     """
     _debug(f"Updating git repository: {repopath}")
     if not repopath.exists():
         raise OSError(f"Can't find path to git repository {repopath}")
+    if check_if_needed and not _git_repo_needs_update(repopath):
+        _debug(f"Repository {repopath} up to date")
+        return
     gitbin = _get_git_binary()
     cwd = os.path.abspath(os.path.curdir)
     os.chdir(str(repopath))
@@ -776,15 +812,22 @@ def _find_system_plugins_path(possible_paths: List[Path]) -> Optional[Path]:
     return None
 
 
+def _load_installation_manifest(path: Path) -> dict:
+    """
+    Load an installation manifest
 
-def _load_manifest(path: str) -> Union[dict, ErrorMsg]:
-    assert os.path.splitext(path)[1] == ".json"
+    An installation manifest is a json file produced during installation,
+    with metadata about what was installed (plugin, name, assets, etc).
+
+    Raises json.JSONDecodeError if the manifest's json could not be parsed
+    """
+    assert path.suffix == '.json'
     try:
         d = json.load(open(path))
         return d
-    except Exception as e:
-        _errormsg(f"Could not parse manifest {path}")
-        return ErrorMsg(str(e))
+    except json.JSONDecodeError as e:
+        _errormsg(f"Could not parse manifest json: {path}")
+        raise e
 
 
 def _is_url(s: str) -> bool:
@@ -821,7 +864,8 @@ def _normalize_version(version: str, default="0.0.0") -> str:
     return ".".join(str(i) for i in versiontup)
 
 
-def _parse_binary(platform: str, binarydef: dict) -> Binary:
+def _parse_binarydef(platform: str, binarydef: dict) -> Binary:
+    assert isinstance(binarydef, dict)
     url = binarydef.get('url')
     if not url:
         raise ParseError(f"Plugin definition for {platform} should have an url")
@@ -832,13 +876,13 @@ def _parse_binary(platform: str, binarydef: dict) -> Binary:
                   extractpath=binarydef.get('extractpath', ''))
 
 
-def _parse_asset(assetdef: dict, defaulturl: str = '') -> Asset:
-    url = assetdef.get('url', defaulturl)
+def _parse_asset(assetdef: dict, defaultsource: str) -> Asset:
+    source = assetdef.get('url', defaultsource)
     extractpath = assetdef.get('extractpath') or assetdef.get('path')
-    if not url and not extractpath:
+    if not source and not extractpath:
         raise ParseError(f"Asset definition should have an URL or an extractpath key")
     paths = extractpath.split(";") if extractpath else None
-    return Asset(url=url, patterns=paths, platform=assetdef.get('platform', 'all'), name=assetdef.get('name', ''))
+    return Asset(source=source, patterns=paths, platform=assetdef.get('platform', 'all'), name=assetdef.get('name', ''))
 
 
 def _enforce_key(d: dict, key: str):
@@ -848,22 +892,37 @@ def _enforce_key(d: dict, key: str):
     return value
 
 
-def _plugin_from_dict(d: dict, pluginurl: str, subpath: str = '') -> Plugin:
+def _plugin_from_dict(d: dict, pluginurl: str, subpath: str) -> Plugin:
+    """
+    Args:
+        d: the loaded json
+        pluginurl: the url of this plugin
+        subpath: the path of the manifest's folder, relative to the root of the repository
+    """
+    clonepath = _git_local_path(pluginurl)
     version = _normalize_version(_enforce_key(d, 'version'))
+    pluginname = _enforce_key(d, 'name')
     opcodes = _enforce_key(d, 'opcodes')
     opcodes.sort()
 
     binaries: Dict[str, Binary] = {}
     for platform, binarydef in _enforce_key(d, 'binaries').items():
         try:
-            binary = _parse_binary(platform, binarydef)
+            binary = _parse_binarydef(platform, binarydef)
             binaries[binary.platform] = binary
         except ParseError as e:
+            _errormsg(f"Failed to parse binary definition for plugin {d.get('name', '??')}")
+            _errormsg(f"... source data: {binarydef}")
             _errormsg(str(e))
 
     if not binaries:
         raise SchemaError("No valid binaries defined")
 
+    manifest_local_folder = clonepath / subpath
+    if not manifest_local_folder.exists():
+        _errormsg(f"The local manifest folder corresponding to plugin {pluginname} was not found "
+                  f"({manifest_local_folder})")
+
     assets: List[Asset] = []
     assetdefs = d.get('assets')
     if assetdefs:
@@ -871,7 +930,7 @@ def _plugin_from_dict(d: dict, pluginurl: str, subpath: str = '') -> Plugin:
             raise SchemaError(f"assets should hold a list of asset definitions, got {assetdefs}")
         for assetdef in assetdefs:
             try:
-                assets.append(_parse_asset(assetdef, defaulturl=pluginurl))
+                assets.append(_parse_asset(assetdef, defaultsource=manifest_local_folder.as_posix()))
             except ParseError as e:
                 _errormsg(str(e))
 
@@ -967,8 +1026,7 @@ def _read_plugindef(filepath: Union[str, Path], url: str = '',
     _debug("... manifest json ok")
 
     try:
-        plugin = _plugin_from_dict(d, pluginurl=url)
-        plugin.manifest_relative_path = manifest_relative_path
+        plugin = _plugin_from_dict(d, pluginurl=url, subpath=manifest_relative_path)
     except SchemaError as e:
         _errormsg(f"Error while processing {filepath}")
         raise e
@@ -986,10 +1044,17 @@ def _normalize_path(path: str) -> str:
     return path
 
 
-def _make_install_manifest(plugin: Plugin, platform: str) -> dict:
+def _make_install_manifest(plugin: Plugin, assetfiles: List[str] = None) -> dict:
     """
     Create an installation manifest dict
+
+    Args:
+        plugin: the Plugin corresponding to this installation manifest
+        assetfiles: if given, a list of asset filenames installed by this plugin
+            (only the filenames, no path: all assets are placed in a flat folder
+            under the plugins prefix)
     """
+    platform = _get_platform()
     out = {}
     out['name'] = plugin.name
     out['author'] = plugin.author
@@ -1001,6 +1066,7 @@ def _make_install_manifest(plugin: Plugin, platform: str) -> dict:
     out['build_platform'] = plugin.binaries[platform].build_platform
     out['binary'] = plugin.binaries[platform].binary_filename()
     out['platform'] = platform
+    out['assetfiles'] = assetfiles or []
     return out
 
 
@@ -1092,11 +1158,10 @@ def system_installed_dlls() -> List[Path]:
     return list(path.glob("*" + ext))
 
 
-def _datarepo_localpath():
-    return RISSET_ROOT / 'risset-data'
-
-
 class MainIndex:
+    """
+    This class holds risset's main index
+    """
     def __init__(self, datarepo: Path = None, update=False):
         """
         Args:
@@ -1121,11 +1186,20 @@ class MainIndex:
         self.version = ''
         self.pluginsources: Dict[str, IndexItem] = {}
         self.plugins: Dict[str, Plugin] = {}
-        self._parseindex(updateindex=updateindex, updateplugins=update)
+        self._cache: Dict[str, Any] = {}
+        self._parse_index(updateindex=updateindex, updateplugins=update)
 
-    def _parseindex(self, updateindex=False, updateplugins=False) -> None:
+    def _parse_index(self, updateindex=False, updateplugins=False, fail_when_debugging=True) -> None:
+        """
+        Parse the main index and each entity defined within it
+
+        If there are errors for a plugin definition, this plugin is skipped and an
+        error message is printed, unless fail_if_error is True, in which case the
+        whole operation is cancelled
+        """
         self.plugins.clear()
         self.pluginsources.clear()
+        self._cache.clear()
         if updateindex:
             _git_update(self.datarepo)
 
@@ -1162,11 +1236,21 @@ class MainIndex:
 
         if self.pluginsources:
             for name, pluginsource in self.pluginsources.items():
-                plugin = self._parse_plugin(name)
-                self.plugins[name] = plugin
+                try:
+                    _debug(f"Parsing plugin definition for {name}")
+                    plugin = self._parse_plugin(name)
+                    self.plugins[name] = plugin
+                except Exception as e:
+                    if register.debug and fail_when_debugging:
+                        raise e
+                    else:
+                        _errormsg(f"Error while parsing plugin definition for {name}: {e}")
 
     def update(self):
-        self._parseindex(updateindex=True, updateplugins=True)
+        """
+        Update all sources and reread the index
+        """
+        self._parse_index(updateindex=True, updateplugins=True, fail_if_error=register.debug)
 
     def build_documentation(self, dest: Path = None, buildhtml=True,  onlyinstalled=False) -> Path:
         """
@@ -1191,11 +1275,10 @@ class MainIndex:
         """
         return _generate_documentation(self, dest=dest, buildhtml=buildhtml, onlyinstalled=onlyinstalled)
 
-    def _parse_plugin(self, pluginname) -> Optional[Plugin]:
+    def _parse_plugin(self, pluginname: str) -> Optional[Plugin]:
         pluginsource = self.pluginsources.get(pluginname)
         if pluginsource is None:
             raise KeyError(f"Plugin {pluginname} not known. Known plugins: {self.pluginsources.keys()}")
-        # todo
         manifestpath = pluginsource.manifest_path()
         assert manifestpath.exists()
         manifeststr = open(manifestpath).read()
@@ -1233,6 +1316,8 @@ class MainIndex:
     def installed_manifests_path(self) -> Path:
         """
         Returns the path to were installation manifests are saved in this system
+
+        Creates the path if it doesn't exist already
         """
         path = RISSET_ROOT / "installed-manifests"
         if not path.exists():
@@ -1298,14 +1383,10 @@ class MainIndex:
                 return None
             return htmlpage
         else:
-            opcodedef: Opcode = self.opcodes.get(opcode)
-            if not opcodedef:
-                raise ValueError(f"Opcode {opcode} not found")
-            plugin = self.plugins.get(opcodedef.plugin)
-            if not plugin:
-                raise RuntimeError(f"The opcode {opcode} is defined in plugin {opcodedef.plugin}"
-                                   f", but the plugin was not found")
-            return plugin.manpage(opcode)
+            for plugin in self.plugins.values():
+                if opcode in plugin.opcodes:
+                    return plugin.manpage(opcode)
+            _errormsg(f"Opcode {opcode} not found")
 
     def installed_plugin_info(self, plugin: Plugin) -> Optional[InstalledPluginInfo]:
         """
@@ -1325,9 +1406,10 @@ class MainIndex:
             pluginkey = manifest.name.split(".")[0]
             name, version = _parse_pluginkey(pluginkey)
             if name == plugin.name:
-                result = _load_manifest(manifest.as_posix())
-                if isinstance(result, ErrorMsg):
-                    _errormsg(str(result))
+                try:
+                    result = _load_installation_manifest(manifest)
+                except Exception as e:
+                    _errormsg(f"Could not load installation manifest for plugin {plugin.name}, skipping")
                     continue
                 installed_version = result['version']
                 installed_manifest_path = manifest
@@ -1418,19 +1500,44 @@ class MainIndex:
         return {opcode.name: opcode
                 for opcode in opcodes}
 
-    def defined_opcodes(self, installed=False) -> List[Opcode]:
+    def opcode_syntaxes(self, opcode: str) -> List[str]:
+        manpage = self.find_manpage(opcode, markdown=True)
+        if not manpage:
+            _errormsg(f"Opcode {opcode} has no manpage")
+            return []
+        text = open(manpage).read()
+        lines = iter(text.splitlines())
+        syntaxlines = []
+        for line in lines:
+            if re.search(r"^\s*#+\s+[s|S]yntax\s*$", line):
+                break
+        else:
+            return []
+        for line in lines:
+            if re.search(r"^\s*[#!;/]", line):
+                break
+            elif opcode in line and (re.search(r"^\s*[akigS\[x]", line) or line.lstrip().startswith(opcode)):
+                syntax = line.strip().split(";", maxsplit=1)[0]
+                syntaxlines.append(syntax)
+        return syntaxlines
+
+    def defined_opcodes(self) -> List[Opcode]:
         """
         Returns a list of opcodes
 
-        Args:
-            installed: only opcodes which are installed are returned
         """
+        cached = self._cache.get('opcodes')
+        if cached:
+            return cached
         opcodes = []
         for plugin in self.plugins.values():
-            if installed and not self.is_plugin_installed(plugin):
-                continue
+            pluginstalled = self.is_plugin_installed(plugin)
+            print(plugin.opcodes)
             for opcodename in plugin.opcodes:
-                opcodes.append(Opcode(name=opcodename, plugin=plugin.name))
+                print("querying ", opcodename)
+                opcodes.append(Opcode(name=opcodename, plugin=plugin.name, syntaxes=self.opcode_syntaxes(opcodename),
+                                      installed=pluginstalled))
+        self._cache['opcodes'] = opcodes
         return opcodes
 
     def match_opcodes(self, globpattern: str, installed=False) -> List[Opcode]:
@@ -1491,29 +1598,28 @@ class MainIndex:
                                 f"{plugin.opcodes[0]}, which is provided by this plugin, "
                                 f"is not present")
 
+        # Install assets, if any
+        assetfiles = []
+        if plugin.assets:
+            for asset in plugin.assets:
+                if asset.platform == 'all' or asset.platform == platform:
+                    _debug(f"Installing asset {asset.identifier()}")
+                    assetfiles.extend(self.install_asset(asset, plugin.name))
+
         # install manifest
-        manifests_path = self.installed_manifests_path()
-        if not manifests_path.exists():
-            manifests_path.mkdir(parents=True)
-        manifest = _make_install_manifest(plugin, platform=platform)
-        manifest_path = manifests_path / f"{plugin.name}.json"
+        manifest_path = self.installed_manifests_path() / f"{plugin.name}.json"
+        manifest = _make_install_manifest(plugin, assetfiles=assetfiles)
         try:
             manifest_json = json.dumps(manifest, indent=True)
         except Exception as e:
-            _errormsg("install_plugin: json error while saving manifest: " + str(e))
+            _errormsg(f"install_plugin: json error while saving manifest: {e}")
+            _errormsg(f"   manifest was: \n{manifest}")
             return ErrorMsg("Error when dumping manifest to json")
 
         with open(manifest_path.as_posix(), "w") as f:
             f.write(manifest_json)
         _debug(f"Saved manifest for plugin {plugin.name} to {manifest_path}")
 
-        # Install assets, if any
-        if plugin.assets:
-            for asset in plugin.assets:
-                if asset.platform == 'all' or asset.platform == platform:
-                    _debug(f"Installing asset {asset.identifier()}")
-                    self.install_asset(asset, plugin.name)
-
         # no errors
         return None
 
@@ -1549,7 +1655,7 @@ class MainIndex:
     def list_plugins(self, installed=False, nameonly=False, allplatforms=False,
                      leftcolwidth=20, oneline=False):
         platform = _get_platform()
-        descr_max_width = 60
+        descr_max_width = os.get_terminal_size().columns - 36
         for plugin in self.plugins.values():
             data = []
             if platform not in plugin.binaries.keys():
@@ -1568,7 +1674,7 @@ class MainIndex:
             extra_lines = []
             if info:
                 if info.versionstr == UNKNOWN_VERSION:
-                    data.append("installed (manually)")
+                    data.append("manual")
                 else:
                     if oneline:
                         data.append(info.versionstr)
@@ -1580,7 +1686,7 @@ class MainIndex:
                 status = "[" + ", ".join(data) + "]"
             else:
                 status = ""
-            leftcol = f"{plugin.name}  @ {plugin.version}"
+            leftcol = f"{plugin.name} /{plugin.version}"
             descr = plugin.short_description
             if oneline and len(descr) > descr_max_width:
                 descr = descr[:descr_max_width] + "…"
@@ -1630,13 +1736,13 @@ class MainIndex:
             print("Assets:")
             for asset in plugdef.assets:
                 print(f"    * identifier: {_abbrev(asset.identifier(), 70)}\n"
-                      f"      url: {asset.url}\n"
-                      f"      paths: {', '.join(asset.patterns)}\n"
+                      f"      source: {asset.source}\n"
+                      f"      patterns: {', '.join(asset.patterns)}\n"
                       f"      platform: {asset.platform}")
         print()
         return True
 
-    def uninstall_plugin(self, plugin: Plugin) -> None:
+    def uninstall_plugin(self, plugin: Plugin, removeassets=True) -> None:
         """
         Uninstall the given plugin
 
@@ -1646,6 +1752,10 @@ class MainIndex:
 
         Raises RuntimeError if the plugin is not installed or is installed
         in the system's folder.
+
+        Args:
+            plugin: the plugin to uninstall
+            removeassets: if True, removes the assets installed by this plugin
         """
         info = self.installed_plugin_info(plugin)
         if not info:
@@ -1659,15 +1769,23 @@ class MainIndex:
         os.remove(info.dllpath.as_posix())
         assert not info.dllpath.exists(), f"Attempted to remove {info.dllpath.as_posix()}, but failed"
         manifestpath = info.installed_manifest_path
+        assetsfolder = RISSET_ASSETS_PATH / plugin.name
         if manifestpath and manifestpath.exists():
+            installed_manifest = _load_installation_manifest(manifestpath)
+            assetfiles = installed_manifest.get('assetfiles', [])
+            if removeassets and assetsfolder.exists():
+                _debug(f"Removing assets for plugin {plugin.name}: {assetfiles}")
+                for assetfile in assetfiles:
+                    assetfullpath = assetsfolder / assetfile
+                    if assetfullpath.exists():
+                        os.remove(assetfullpath)
+                remainingassets = list(assetsfolder.glob("*"))
+                if remainingassets:
+                    _info(f"There are remaining assets in the folder {assetsfolder}: {remainingassets}")
+                    _info("... They will be removed")
+                _rm_dir(assetsfolder)
             os.remove(manifestpath.as_posix())
 
-    def asset_folder(self, prefix: str = '') -> path:
-        if not prefix:
-            return RISSET_ASSETS_PATH
-        else:
-            return RISSET_ASSETS_PATH / prefix
-
     def install_asset(self, asset: Asset, prefix: str) -> List[str]:
         """
         Install an Asset under a given prefix
@@ -1680,7 +1798,7 @@ class MainIndex:
             a list of installed file names (only the filename, not the
             absolute path, the destination path is RISSET_ASSETS_PATH / prefix)
         """
-        destination_folder = self.asset_folder(prefix)
+        destination_folder = RISSET_ASSETS_PATH / prefix
         sources = asset.retrieve()
         if sources:
             destination_folder.mkdir(parents=True, exist_ok=True)
@@ -1692,35 +1810,53 @@ class MainIndex:
                 shutil.copy(source, destination_folder)
         return [f.name for f in sources]
 
+    def generate_opcodes_xml(self) -> str:
+        """
+        Generates xml following the scheme of the manual's opcodes.xml
+
+        This can be used by frontends to generate help for all csound opcodes
+        """
+        lines =  []
+        _ = lines.append
+        _('<?xml version="1.0" encoding="UTF-8"?>')
+        _('<opcodes>')
+        # For now, gather all opcodes belonging to one plugin under the same category. Later we can
+        # enforce that each plugin defines a category in their manpage
+        for plugin in self.plugins.values():
+            _(f'<category name="External Plugin:{plugin.name}')
+            for opcode in plugin.opcodes:
+                pass
+            _('</category')
+
+        _('</opcodes>')
+
+
 
 ###############################################################
 #                        Documentation                        #
 ###############################################################
 
 
+def _is_package_installed(pkg: str) -> bool:
+    import importlib.util
+    return importlib.util.find_spec(pkg) is not None
+
+
 def _call_mkdocs(folder: Path, *args: str):
     currentdir = os.getcwd()
     os.chdir(folder)
-    subprocess.call(["python", "-m", "mkdocs"] + list(args))
+    subprocess.call([sys.executable, "-m", "mkdocs"] + list(args))
     os.chdir(currentdir)
 
 
 def _is_mkdocs_installed() -> bool:
-    if sys.platform == "linux" or sys.platform == "darwin":
-        return shutil.which("mkdocs") is not None
-    else:
-        try:
-            import mkdocs
-            return True
-        except ImportError as e:
-            return False
+    return _is_package_installed("mkdocs") or shutil.which("mkdocs") is not None
 
 
 def _generate_documentation(index: MainIndex, dest: Path = None, 
                             buildhtml=True, onlyinstalled=False
                             ) -> Path:
     if dest is None:
-        # ~/.local/share/risset/man
         dest = RISSET_GENERATED_DOCS
     _compile_docs(index=index, dest=dest / "docs", makeindex=True,
                   onlyinstalled=onlyinstalled)
@@ -1738,7 +1874,7 @@ def _generate_documentation(index: MainIndex, dest: Path = None,
 def _compile_docs(index: MainIndex, dest: Path, makeindex=True,
                   onlyinstalled=False) -> None:
     """
-    Gathers all manpages and generates a mkdocs compatible docs folder
+    Gather all manpages and generate a mkdocs compatible docs folder
     """
     dest = dest.expanduser().absolute()
     if dest.exists():
@@ -1874,7 +2010,6 @@ def cmd_list(mainindex: MainIndex, args) -> None:
     """
     Lists all plugins available for download
     """
-    # TODO: implement flags: --json, --output
     if args.json:
         d = mainindex._list_plugins_as_dict(installed=args.installed, allplatforms=args.all)
         if args.outfile:
@@ -2032,28 +2167,23 @@ def cmd_man(idx: MainIndex, args) -> bool:
     return True
 
 
-def cmd_update(idx: MainIndex, args) -> bool:
-    idx.update()
-    return True
-
-
 def cmd_resetcache(args) -> None:
     _rm_dir(RISSET_DATAREPO_LOCALPATH)
     _rm_dir(RISSET_CLONES_PATH)
 
 
 def update_self():
-    """ upgrade risset itself """
-    python = sys.executable
+    """Upgrade risset itself"""
     _info("Updating risset")
-    subprocess.check_call([python, "-m", "pip", "install", "risset", "--upgrade"])
+    subprocess.check_call([sys.executable, "-m", "pip", "install", "risset", "--upgrade"])
 
 
 def cmd_list_installed_opcodes(plugins_index: MainIndex, args) -> bool:
     """
     Print a list of installed opcodes
+
     """
-    opcodes = plugins_index.defined_opcodes(installed=True)
+    opcodes = plugins_index.defined_opcodes()
     opcodes.sort(key=lambda opcode: opcode.name.lower())
     for opcode in opcodes:
         print(opcode.name)
@@ -2084,12 +2214,6 @@ def _print_file(path: Path) -> None:
     print(text)
 
 
-def _has(cmd: str) -> bool:
-    """ Returns True if cmd is in the path """
-    path = shutil.which(cmd)
-    return path is not None
-
-
 def _show_markdown_file(path: Path) -> None:
     if not _running_from_terminal():
         _open_in_default_application(str(path))
@@ -2102,119 +2226,86 @@ def _show_markdown_file(path: Path) -> None:
     print(highlight(code, MarkdownLexer(), TerminalTrueColorFormatter()))
 
 
-def _show_markdown_file_(path: Path) -> None:
-    if not _running_from_terminal():
-        _open_in_default_application(str(path))
-        return
-
-    if sys.platform == 'linux' or sys.platform == 'darwin':
-        if _has("bat"):
-            subprocess.call(["bat", "--style", "header", str(path)])
-        elif _has("pygmentize"):
-            subprocess.call(["pygmentize", str(path)])
-        elif _has("highlight"):
-            #  highlight --force --out-format="${highlight_format}" --style="${HIGHLIGHT_STYLE}"
-            subprocess.call(["highlight", "--force", "--out-format=ANSI", str(path)])
-        else:
-            print("Considere installing 'pygmentize', 'highlight' or 'bat' for better syntax highlighting")
-            print()
-            _print_file(path)
-    else:
-        if _has("bat"):
-            subprocess.call(["bat", "--style", "header", str(path)], shell=True)
-        elif _has("pygmentize"):
-            subprocess.call(["pygmentize", str(path)])
-        elif _has("highlight"):
-            #  highlight --force --out-format="${highlight_format}" --style="${HIGHLIGHT_STYLE}"
-            subprocess.call(["highlight", "--force", "--out-format=ANSI", str(path)])
-        else:
-            print("Considere installing 'highlight', 'bat' or 'pygmentize' for better syntax highlighting")
-            print()
-            _print_file(path)
-
-
-def add_flag(parser, flag, help=""):
-    parser.add_argument(flag, action="store_true", help=help)
-
-
 def main():
     # Preliminary checks
     if sys.platform not in ("linux", "darwin", "win32"):
         _errormsg(f"Platform not supported: {sys.platform}")
         sys.exit(-1)
 
-    if _get_binary("git") is None:
+    if _get_git_binary() is None:
         _errormsg("git command not found. Check that git is installed and in the PATH")
         sys.exit(-1)
 
+    def flag(parser, flag, help=""):
+        parser.add_argument(flag, action="store_true", help=help)
+
     # Main parser
     parser = argparse.ArgumentParser()
-    add_flag(parser, "--debug", help="Print debug information")
-    add_flag(parser, "--update", help="Update the plugins data before any action")
-    add_flag(parser, "--version")
+    flag(parser, "--debug", help="Print debug information")
+    flag(parser, "--update", help="Update the plugins data before any action")
+    flag(parser, "--version")
     subparsers = parser.add_subparsers(dest='command')
 
     # List command
-    list_group = subparsers.add_parser('list', help="List packages")
-    add_flag(list_group, "--json", help="Outputs list as json")
-    add_flag(list_group, "--all", "List all plugins, even those without a binary for the current platform")
-    add_flag(list_group, "--nameonly", help="Output just the name of each plugin")
-    add_flag(list_group, "--installed", help="List only installed plugins")
-    add_flag(list_group, "--notinstalled", help="List only plugins which are not installed")
-    list_group.add_argument("-o", "--outfile", help="Outputs to a file")
-    list_group.add_argument("-1", "--oneline", action="store_true", help="List each plugin in one line")
-    list_group.set_defaults(func=cmd_list)
+    list_cmd = subparsers.add_parser('list', help="List packages")
+    flag(list_cmd, "--json", help="Outputs list as json")
+    flag(list_cmd, "--all", "List all plugins, even those without a binary for the current platform")
+    flag(list_cmd, "--nameonly", help="Output just the name of each plugin")
+    flag(list_cmd, "--installed", help="List only installed plugins")
+    flag(list_cmd, "--notinstalled", help="List only plugins which are not installed")
+    list_cmd.add_argument("-o", "--outfile", help="Outputs to a file")
+    list_cmd.add_argument("-1", "--oneline", action="store_true", help="List each plugin in one line")
+    list_cmd.set_defaults(func=cmd_list)
 
     # Install command
-    install_group = subparsers.add_parser("install", help="Install a package")
-    add_flag(install_group, "--force", help="Force install/reinstall")
-    install_group.add_argument("plugins", nargs="+",
-                               help="Name of the plugin/plugins to install. "
-                                    "Glob pattern are supported (enclose them inside quotation marks)")
-    install_group.set_defaults(func=cmd_install)
+    install_cmd = subparsers.add_parser("install", help="Install a package")
+    flag(install_cmd, "--force", help="Force install/reinstall")
+    install_cmd.add_argument("plugins", nargs="+",
+                             help="Name of the plugin/plugins to install. "
+                                  "Glob pattern are supported (enclose them inside quotation marks)")
+    install_cmd.set_defaults(func=cmd_install)
 
     # remove command
-    rm_group = subparsers.add_parser("remove", help="Remove a package")
-    rm_group.add_argument("plugin", nargs="+", help="Plugin/s to remove")
-    rm_group.set_defaults(func=cmd_rm)
+    rm_cmd = subparsers.add_parser("remove", help="Remove a package")
+    rm_cmd.add_argument("plugin", nargs="+", help="Plugin/s to remove")
+    rm_cmd.set_defaults(func=cmd_rm)
 
     # show command
-    show_group = subparsers.add_parser("show", help="Show information about a plugin")
-    show_group.add_argument("plugin", help="Plugin to gather information about")
-    show_group.set_defaults(func=cmd_show)
+    show_cmd = subparsers.add_parser("show", help="Show information about a plugin")
+    show_cmd.add_argument("plugin", help="Plugin to gather information about")
+    show_cmd.set_defaults(func=cmd_show)
 
     # build docs
-    makedocs_group = subparsers.add_parser("makedocs", help="Build the documentation for all defined plugins. "
+    makedocs_cmd = subparsers.add_parser("makedocs", help="Build the documentation for all defined plugins. "
                                                             "This depends on mkdocs being installed")
-    makedocs_group.add_argument("--onlyinstalled", action="store_true", help="Build docs only for installed plugins")
-    makedocs_group.add_argument("-o", "--outfolder", help="Destination folder to place the documentation",
+    makedocs_cmd.add_argument("--onlyinstalled", action="store_true", help="Build docs only for installed plugins")
+    makedocs_cmd.add_argument("-o", "--outfolder", help="Destination folder to place the documentation",
                                 default='')
-    makedocs_group.set_defaults(func=cmd_makedocs)
+    makedocs_cmd.set_defaults(func=cmd_makedocs)
 
     # man command
-    man_group = subparsers.add_parser("man", help="Open manual page for an installed opcode. "
-                                                  "Multiple opcodes or a glob wildcard are allowed")
-    man_group.add_argument("-p", "--path", action="store_true",
-                           help="Only print the path of the manual page. The format is <opcode>:<path>")
-    man_group.add_argument("-s", "--simplepath", action="store_true",
-                           help="Print just the path of the manual page")
-    man_group.add_argument("-m", "--markdown", action="store_true",
-                           help="Use the .md page instead of the .html version")
-    man_group.add_argument("-e", "--external", action="store_true",
-                           help="Open the man page in the default app. This is only"
+    man_cmd = subparsers.add_parser("man", help="Open manual page for an installed opcode. "
+                                                "Multiple opcodes or a glob wildcard are allowed")
+    man_cmd.add_argument("-p", "--path", action="store_true",
+                         help="Only print the path of the manual page. The format is <opcode>:<path>")
+    man_cmd.add_argument("-s", "--simplepath", action="store_true",
+                         help="Print just the path of the manual page")
+    man_cmd.add_argument("-m", "--markdown", action="store_true",
+                         help="Use the .md page instead of the .html version")
+    man_cmd.add_argument("-e", "--external", action="store_true",
+                         help="Open the man page in the default app. This is only"
                                 " used when opening the markdown man page.")
-    man_group.add_argument("--html", action="store_true",
-                           help="Use the .html page (opens it in the default browser")
-    man_group.add_argument("opcode", nargs="*", help="Show the manual page of this opcode/opcodes")
-    man_group.set_defaults(func=cmd_man)
+    man_cmd.add_argument("--html", action="store_true",
+                         help="Use the .html page (opens it in the default browser")
+    man_cmd.add_argument("opcode", nargs="*", help="Show the manual page of this opcode/opcodes")
+    man_cmd.set_defaults(func=cmd_man)
 
     # update command
-    update_group = subparsers.add_parser("update", help="Update repository")
-    update_group.set_defaults(func=cmd_update)
+    update_cmd = subparsers.add_parser("update", help="Update repository")
 
     # list-opcodes
-    listopcodes = subparsers.add_parser("listopcodes", help="List installed opcodes")
-    listopcodes.set_defaults(func=cmd_list_installed_opcodes)
+    listopcode_cmd = subparsers.add_parser("listopcodes", help="List installed opcodes")
+    listopcode_cmd.set_defaults(func=cmd_list_installed_opcodes)
     
     # reset
     resetgroup = subparsers.add_parser("resetcache", help="Remove local clones of plugin's repositories")
@@ -2232,7 +2323,7 @@ def main():
 
     try:
         _debug("Creating main index")
-        mainindex = MainIndex(update=args.update or args.command == 'update')
+        mainindex = MainIndex(update=(args.update or args.command == 'update'))
     except Exception as e:
         _debug("Failed to create main index")
         if register.debug:
